from eip712_structs import (Address, Boolean, EIP712Struct, Int, Uint,
                            make_domain)
from eth_utils import keccak
from hexbytes import HexBytes

from hubble_exchange.constants import CHAIN_ID, OrderBookContractAddress
from hubble_exchange.models import Order as OrderModel

domain = make_domain(name='Hubble', version="2.0", chainId=CHAIN_ID, verifyingContract=OrderBookContractAddress)
domain_hash = HexBytes(domain.hash_struct())


# Class name must match the struct name in the solidity contract
class Order(EIP712Struct):
    ammIndex = Uint(256)
    trader = Address()
    baseAssetQuantity = Int(256)
    price = Uint(256)
    salt = Uint(256)
    reduceOnly = Boolean()


def get_order_hash(order: OrderModel) -> HexBytes:
    order_struct = Order(
        ammIndex=order.amm_index,
        trader=order.trader,
        baseAssetQuantity=order.base_asset_quantity,
        price=order.price,
        salt=order.salt,
        reduceOnly=order.reduce_only,
    )

    order_struct_hash = HexBytes(order_struct.hash_struct())
    order_hash_bytes = b'\x19\x01' + domain_hash + order_struct_hash
    order_hash = HexBytes(keccak(order_hash_bytes))
    return order_hash



import json
from typing import Any, Callable, List

import websocket
from hexbytes import HexBytes

from hubble_exchange.eth import (get_web3_client,
                                 get_websocket_endpoint)
from hubble_exchange.models import (GetPositionsResponse, Order,
                                    OrderBookDepthResponse,
                                    OrderBookDepthUpdateResponse,
                                    OrderStatusResponse, WebsocketResponse)
from hubble_exchange.order_book import OrderBookClient
from hubble_exchange.utils import (float_to_scaled_int,
                                   get_address_from_private_key, get_new_salt)


class HubbleClient:
    def __init__(self, private_key: str):
        if not private_key:
            raise ValueError("Private key is not set")
        self.trader_address = get_address_from_private_key(private_key)
        if not self.trader_address:
            raise ValueError("Cannot determine trader address from private key")

        self.web3_client = get_web3_client()
        self.websocket_endpoint = get_websocket_endpoint()
        self.order_book_client = OrderBookClient(private_key)

    def get_order_book(self, market: int) -> OrderBookDepthResponse:
        return self.web3_client.eth.get_order_book_depth(market)

    def get_margin_and_positions(self) -> GetPositionsResponse:
        return self.web3_client.eth.get_margin_and_positions(self.trader_address)

    def get_order_status(self, order_id: HexBytes) -> OrderStatusResponse:
        return self.web3_client.eth.get_order_status(order_id.hex())

    def place_orders(self, orders: List[Order], tx_options = None, mode=None) -> List[Order]:
        if len(orders) > 75:
            raise ValueError("Cannot place more than 75 orders at once")

        for order in orders:
            if order.amm_index is None:
                raise ValueError("Order AMM index is not set")
            if order.base_asset_quantity is None:
                raise ValueError("Order base asset quantity is not set")
            if order.price is None:
                raise ValueError("Order price is not set")
            if order.reduce_only is None:
                raise ValueError("Order reduce only is not set")

            # trader and salt can be set automatically
            if order.trader in [None, "0x", ""]:
                order.trader = self.trader_address
            if order.salt in [None, 0]:
                order.salt = get_new_salt()

        return self.order_book_client.place_orders(orders, tx_options, mode)

    def place_single_order(
        self, market: int, base_asset_quantity: float, price: float, reduce_only: bool, tx_options = None, mode=None
    ) -> Order:
        order = Order(
            id=None,
            amm_index=market,
            trader=self.trader_address,
            base_asset_quantity=float_to_scaled_int(base_asset_quantity, 18),
            price=float_to_scaled_int(price, 6),
            salt=get_new_salt(),
            reduce_only=reduce_only,
        )
        order_hash = self.order_book_client.place_order(order, tx_options, mode)
        order.id = order_hash
        return order

    def cancel_orders(self, orders: List[Order], tx_options = None, mode=None) -> None:
        if len(orders) > 100:
            raise ValueError("Cannot cancel more than 100 orders at once")

        self.order_book_client.cancel_orders(orders, tx_options, mode)

    def cancel_order_by_id(self, order_id: HexBytes, tx_options = None, mode=None) -> None:
        order_status = self.get_order_status(order_id)
        position_side_multiplier = 1 if order_status.positionSide == "LONG" else -1
        order = Order(
            id=order_id.hex(),
            amm_index=order_status.symbol,
            trader=self.trader_address,
            base_asset_quantity=float_to_scaled_int(float(order_status.origQty) * position_side_multiplier, 18),
            price=float_to_scaled_int(float(order_status.price), 6),
            salt=int(order_status.salt),
            reduce_only=order_status.reduceOnly,
        )
        self.cancel_orders([order], tx_options, mode)

    def subscribe_to_order_book_depth(
        self, market: int, callback: Callable[[websocket.WebSocketApp, OrderBookDepthUpdateResponse], Any]
    ) -> None:

        def on_open(ws):
            msg = {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "trading_subscribe",
                "params": ["streamDepthUpdateForMarket", market]
            }
            ws.send(json.dumps(msg))

        def on_message(ws, message):
            message_json = json.loads(message)
            response = WebsocketResponse(**message_json)
            if response.method and response.method == "trading_subscription":
                response = OrderBookDepthUpdateResponse(
                    T=response.params['result']['T'],
                    symbol=response.params['result']['s'],
                    bids=response.params['result']['b'],
                    asks=response.params['result']['a'],
                )
                callback(ws, response)

        ws = websocket.WebSocketApp(self.websocket_endpoint, on_open=on_open, on_message=on_message)
        ws.run_forever()



import math
import random
import time

from eth_account import Account
from eth_typing import Address


def float_to_scaled_int(val: float, scale: int) -> int:
    big_float = math.floor(val * (10 ** scale))
    return int(big_float)


def get_address_from_private_key(private_key: str) -> Address:
    account = Account.from_key(private_key)
    return account.address

def get_new_salt() -> int:
    return int(str(time.time_ns()) + str(random.randint(0, 10000)))



import json
import os
from enum import Enum
from typing import Any, Dict, List

from hexbytes import HexBytes

from hubble_exchange.constants import (CHAIN_ID, GAS_PER_ORDER, MAX_GAS_LIMIT,
                                       OrderBookContractAddress)
from hubble_exchange.eip712 import get_order_hash
from hubble_exchange.eth import HubblenetWeb3 as Web3
from hubble_exchange.eth import get_web3_client
from hubble_exchange.models import Order
from hubble_exchange.utils import get_address_from_private_key

# read abi from file
HERE = os.path.dirname(__file__)
with open(f"{HERE}/contract_abis/OrderBook.json", 'r') as abi_file:
    abi_str = abi_file.read()
    ABI = json.loads(abi_str)


class TransactionMode(Enum):
    no_wait = 0
    wait_for_accept = 1
    wait_for_head = 1


class OrderBookClient(object):
    def __init__(self, private_key: str):
        self._private_key = private_key
        self.public_address = get_address_from_private_key(private_key)

        self.web3_client = get_web3_client()
        self.order_book = self.web3_client.eth.contract(address=OrderBookContractAddress, abi=ABI)
        self.nonce = self.web3_client.eth.get_transaction_count(self.public_address)
        self.transaction_mode = TransactionMode.no_wait

    def set_transaction_mode(self, mode: TransactionMode):
        self.transaction_mode = mode

    def place_order(self, order: Order, custom_tx_options=None, mode=None) -> HexBytes:
        order_hash = get_order_hash(order)

        tx_options = {'gas': GAS_PER_ORDER}
        tx_options.update(custom_tx_options or {})

        self._send_orderbook_transaction("placeOrder", [order.to_dict()], tx_options, mode)
        return order_hash

    def place_orders(self, orders: List[Order], custom_tx_options=None, mode=None) -> List[Order]:
        """
        Place multiple orders at once. This is more efficient than placing them one by one.
        """
        place_order_payload = []

        for order in orders:
            order_hash = get_order_hash(order)
            order.id = order_hash
            place_order_payload.append(order.to_dict())

        tx_options = {'gas': min(GAS_PER_ORDER * len(orders), MAX_GAS_LIMIT)}
        tx_options.update(custom_tx_options or {})
        self._send_orderbook_transaction("placeOrders", [place_order_payload], tx_options, mode)
        return orders

    def cancel_orders(self, orders: list[Order], custom_tx_options=None, mode=None) -> None:
        cancel_order_payload = []
        for order in orders:
            cancel_order_payload.append(order.to_dict())

        tx_options = {'gas': min(GAS_PER_ORDER * len(orders), MAX_GAS_LIMIT)}
        tx_options.update(custom_tx_options or {})

        self._send_orderbook_transaction("cancelOrders", [cancel_order_payload], tx_options, mode)

    def _get_nonce(self) -> int:
        if self.nonce is None:
            self.nonce = self.web3_client.eth.get_transaction_count(self.public_address)
        else:
            self.nonce += 1
        return self.nonce - 1

    def _send_orderbook_transaction(self, method_name: str, args: List[Any], tx_options: Dict, mode: TransactionMode) -> HexBytes:
        if mode is None:
            mode = self.transaction_mode

        method = getattr(self.order_book.functions, method_name)
        nonce = self._get_nonce()
        tx_params = {
            'from': self.public_address,
            'chainId': CHAIN_ID,
            'maxFeePerGas': Web3.to_wei(60, 'gwei'),  # base + tip
            'maxPriorityFeePerGas': 0,  # tip
            'nonce': nonce,
        }
        if tx_options:
            tx_params.update(tx_options)

        transaction = method(*args).build_transaction(tx_params)
        signed_tx = self.web3_client.eth.account.sign_transaction(transaction, self._private_key)
        tx_hash = self.web3_client.eth.send_raw_transaction(signed_tx.rawTransaction)
        if mode == TransactionMode.wait_for_accept:
            self._wait_for_accept(tx_hash)
        elif mode == TransactionMode.wait_for_head:
            self._wait_for_head(tx_hash)

        return tx_hash

    def _wait_for_accept(self, tx_hash: HexBytes):
        self.web3_client.eth.wait_for_transaction_receipt(tx_hash, timeout=120, poll_latency=0.1)

    def _wait_for_head(self, tx_hash: HexBytes):
        self.web3_client.eth.wait_for_transaction_status(tx_hash, timeout=120, poll_latency=0.1)




import os
from typing import Callable, Dict

from eth_typing import Address
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3.eth.eth import Eth, Timeout
from web3.exceptions import TimeExhausted
from web3.main import Web3, get_default_modules
from web3.method import Method, default_root_munger
from web3.middleware import geth_poa_middleware
from web3.middleware.cache import construct_simple_cache_middleware
from web3.types import RPCEndpoint, _Hash32

from hubble_exchange.errors import OrderNotFound, TraderNotFound
from hubble_exchange.models import (GetPositionsResponse,
                                    OrderBookDepthResponse,
                                    OrderStatusResponse)


rpc_endpoint = ""
websocket_endpoint = ""


class HubblenetEth(Eth):
    _get_transaction_status: Method[Callable[[_Hash32], Dict]] = Method(RPCEndpoint("eth_getTransactionStatus"), mungers=[default_root_munger])
    _get_order_status: Method[Callable[[_Hash32], Dict]] = Method(RPCEndpoint("trading_getOrderStatus"), mungers=[default_root_munger])
    _get_margin_and_positions: Method[Callable[[Address], Dict]] = Method(RPCEndpoint("trading_getMarginAndPositions"), mungers=[default_root_munger])
    _get_order_book_depth: Method[Callable[[int], Dict]] = Method(RPCEndpoint("trading_getTradingOrderBookDepth"), mungers=[default_root_munger])

    def get_transaction_status(self, transaction_hash: _Hash32) -> Dict:
        return self._get_transaction_status(transaction_hash)

    def get_order_status(self, order_id: _Hash32) -> OrderStatusResponse:
        try:
            response = self._get_order_status(order_id)
            return OrderStatusResponse(**response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "order not found":
                raise OrderNotFound()
            else:
                raise e

    def get_margin_and_positions(self, trader: Address) -> GetPositionsResponse:
        try:
            response = self._get_margin_and_positions(trader)
            return GetPositionsResponse(**response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "trader not found":
                raise TraderNotFound()
            else:
                raise e

    def get_order_book_depth(self, market: int) -> OrderBookDepthResponse:
        response = self._get_order_book_depth(market)
        return OrderBookDepthResponse(**response)

    def wait_for_transaction_status(self, transaction_hash: HexBytes, timeout: float = 120, poll_latency: float = 0.1) -> Dict:
        try:
            with Timeout(timeout) as _timeout:
                while True:
                    try:
                        tx_status = self._get_transaction_status(transaction_hash.hex())
                    except:
                        tx_status = None
                    if tx_status is None or tx_status['status'] == "NOT_FOUND":
                        _timeout.sleep(poll_latency)
                        continue
                    else:
                        break
            return tx_status

        except Timeout:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not in the chain "
                f"after {timeout} seconds"
            )


def get_web3_modules() -> Dict:
    modules = get_default_modules()
    modules["eth"] = HubblenetEth
    return modules

class HubblenetWeb3(Web3):
    eth: HubblenetEth


def get_rpc_endpoint() -> str:
    global rpc_endpoint
    if not rpc_endpoint:
        rpc_host = os.getenv("HUBBLE_RPC_HOST")
        if not rpc_host:
            raise ValueError("HUBBLE_RPC_HOST environment variable not set")
        blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
        if not blockchain_id:
            raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
        path = f"/ext/bc/{blockchain_id}/rpc"
        rpc_endpoint = f"https://{rpc_host}{path}"
    return rpc_endpoint


def get_websocket_endpoint() -> str:
    global websocket_endpoint
    if not websocket_endpoint:
        rpc_host = os.getenv("HUBBLE_RPC_HOST")
        if not rpc_host:
            raise ValueError("HUBBLE_RPC_HOST environment variable not set")
        blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
        if not blockchain_id:
            raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
        path = f"/ext/bc/{blockchain_id}/ws"
        websocket_endpoint = f"wss://{rpc_host}{path}"
    return websocket_endpoint


def get_web3_client() -> HubblenetWeb3:
    rpc_endpoint = get_rpc_endpoint()
    
    web3_client = HubblenetWeb3(HTTPProvider(rpc_endpoint), modules=get_web3_modules())
    web3_client.middleware_onion.inject(geth_poa_middleware, layer=0)

    # cache frequent eth_chainId calls
    cache_chain_id_middleware = construct_simple_cache_middleware()
    web3_client.middleware_onion.add(cache_chain_id_middleware, name="cache")
    return web3_client



from hubble_exchange.client import HubbleClient
from hubble_exchange.models import (GetPositionsResponse, Order,
                                    OrderBookDepthResponse,
                                    OrderBookDepthUpdateResponse,
                                    OrderStatusResponse)
from hubble_exchange.order_book import TransactionMode



from dataclasses import dataclass
from typing import List

from eth_typing import Address
from hexbytes import HexBytes

from hubble_exchange.utils import float_to_scaled_int, get_new_salt


@dataclass
class Order:
    id: HexBytes
    amm_index: int
    trader: Address
    base_asset_quantity: int
    price: int
    salt: int
    reduce_only: bool

    def to_dict(self):
        return {
            "ammIndex": self.amm_index,
            "trader": self.trader,
            "baseAssetQuantity": self.base_asset_quantity,
            "price": self.price,
            "salt": self.salt,
            "reduceOnly": self.reduce_only,
        }

    @classmethod
    def new(cls, amm_index: int, base_asset_quantity: float, price: float, reduce_only: bool):
        """
        Create a new order with a random salt and no ID or trader. This can be used for placing
        multiple orders at once.
        """
        return cls(
            id=None,
            amm_index=amm_index,
            trader=None,
            base_asset_quantity=float_to_scaled_int(base_asset_quantity, 18),
            price=float_to_scaled_int(price, 6),
            salt=get_new_salt(),
            reduce_only=reduce_only)


@dataclass
class OrderStatusResponse:
    executedQty: str
    orderId: str
    origQty: str
    price: str
    reduceOnly: bool
    positionSide: str
    status: str
    symbol: int
    time: int
    type: str
    updateTime: int
    salt: int


@dataclass
class OrderBookDepthResponse:
    lastUpdateId: int
    E: int
    T: int
    symbol: int
    bids: List[List[str]]
    asks: List[List[str]]


class Market(int):
    pass


@dataclass
class Position:
    market: Market
    openNotional: str
    size: str
    unrealisedFunding: str
    liquidationThreshold: str
    notionalPosition: str
    unrealisedProfit: str
    marginFraction: str
    liquidationPrice: str
    markPrice: str


@dataclass
class GetPositionsResponse:
    margin: str
    reservedMargin: str
    positions: List[Position]


@dataclass
class Message:
    jsonrpc: str
    id: int
    method: str
    params: List[any]


@dataclass
class Params:
    subscription: str
    result: any


@dataclass
class WebsocketResponse:
    jsonrpc: str
    method: str
    params: Params


@dataclass
class OrderBookDepthUpdateResponse:
    T: int
    symbol: int
    bids: List[List[str]]
    asks: List[List[str]]



from eth_typing import Address

OrderBookContractAddress = Address("0x0300000000000000000000000000000000000000")
CHAIN_ID = 321123
MAX_GAS_LIMIT = 7_000_000  # 7 million
GAS_PER_ORDER = 200_000  # 200k
